
import pandas as pd
import numpy as np


def load_database():
    # set up to run the R code
    try:
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
        import anndata2ri
        anndata2ri.activate()
        robjects = robjects
        anndata2ri = anndata2ri

        cellchat = importr("CellChat")
    except ImportError:
        raise ImportError("Ensure rpy2 and CellChat are installed correctly.")

    cellchat_db = robjects.r['CellChatDB.mouse'][0]
    interaction_cellchat = pd.DataFrame(cellchat_db).fillna('').transpose()
    interaction_cellchat.columns = list(cellchat_db.colnames)

    # Restructure column names to desired format
    is_secreted = interaction_cellchat["annotation"] == "Secreted Signaling"
    not_ecm = np.logical_not(interaction_cellchat["annotation"] == "ECM-Receptor")
    interaction_cellchat = interaction_cellchat[["interaction_name", "interaction_name_2", "ligand", "receptor",
                                                 "pathway_name"]].copy()
    interaction_cellchat["secreted"] = is_secreted

    interaction_cellchat = interaction_cellchat[not_ecm].reset_index(drop=True)
    interaction = interaction_cellchat

    return interaction


if __name__=="__main__":
    interaction = load_database()
    print(interaction)


import scanpy as sc
import pandas as pd
import anndata

def get_activity_dataframe(adata):
    obs_names = list(adata.obs)
    activity_names = [v for v in obs_names if "activity_" in v]
    return sc.get.obs_df(adata, keys=activity_names)


def write(adata, name):
    adata = adata.copy()
    adata.obs.to_pickle(f'{name}.pkl')
    adata.obs = pd.DataFrame(index=adata.obs.index)
    adata.write(f'{name}.h5ad')


def read(name):
    adata = anndata.read_h5ad(f'{name}.h5ad')
    adata.obs = pd.read_pickle(f'{name}.pkl')
    return adata


from sklearn.preprocessing import normalize
from sklearn.cluster import SpectralClustering, DBSCAN, OPTICS
from skimage.filters.thresholding import threshold_otsu
from statsmodels.stats.multitest import fdrcorrection
from scipy.stats import mannwhitneyu
from scipy.spatial import Delaunay
from tqdm import tqdm
from shapely import geometry
from shapely.ops import unary_union, polygonize
from multiprocessing import Pool
import matplotlib as mpl

from geospace.cci.spatial import get_neighbor_adjacency

import pandas as pd
import numpy as np


def compute_gene_hotspots(adata, gene_list=None, verbose=False, return_hotspot_counts=False,
                          log=False, **kwargs):
    if gene_list is None:
        gene_list = adata.var_names

    num_hotspot_genes = 0

    # gene_list = gene_list[:max_genes]

    data = adata[:, gene_list].X.tocsc()

    region_dict = {}
    hotspot_counts = {}
    if verbose:
        gene_list_range = tqdm(enumerate(gene_list), total=len(gene_list))
    else:
        gene_list_range = enumerate(gene_list)
    for idx, gene in gene_list_range:
        gene_expression = data[:, idx]
        try:
            gene_expression = gene_expression.toarray()
        except AttributeError:
            pass

        cutoff = _compute_cutoff(gene_expression, log=log)
        # cutoff = np.quantile(gene_expression, expression_threshold)
        inds = np.where(gene_expression > cutoff)[0]

        if len(inds) == 0:
            continue

        regions = compute_hotspots(adata, input_data=inds,  # output_key=f"hotspots_{gene}",
                                   output_key=None, return_regions=True,
                                   input_threshold=cutoff, **kwargs)

        if regions is not None:
            num_hotspot_genes += 1
            region_dict[f"hotspots_{gene}"] = regions
            hotspot_counts[gene] = len(regions.categories)

        # if verbose:
        #    print(f"{gene} ({idx}/{len(gene_list)}, {counter} hotspot genes found).")

    hotspots_df = pd.DataFrame(region_dict, index=adata.obs.index)
    adata.obs = pd.concat([adata.obs, hotspots_df], axis=1)

    if return_hotspot_counts:
        return hotspot_counts
    else:
        return num_hotspot_genes


def compute_hotspots(adata, min_samples, eps, input_key=None, output_key=None,
                     input_data=None, return_regions=False,
                     min_size=None,
                     input_threshold=None, core_only=True, method=None):
    if input_data is not None:
        inds = input_data
    else:
        # handle either obs or gene
        try:
            v = adata.obs[input_key]
        except KeyError:
            v = adata[:, input_key].X
            try:
                v = v.toarray()
            except AttributeError:
                pass

        # handle continuous data with provided threshold
        if input_threshold is not None:
            v = v > input_threshold

        inds = np.where(v)[0]

    coords = adata.obsm['spatial']
    X = coords[inds]

    if X.shape[0] < min_size:
        return None

    if method is None or method == "DBSCAN":
        clusters = DBSCAN(min_samples=min_samples, eps=eps)
    elif method == "OPTICS":
        clusters = OPTICS(min_samples=min_samples, max_eps=eps, cluster_method="xi")
    else:
        raise NotImplementedError
    clusters.fit(X)
    labels = clusters.labels_

    if core_only:
        labels_ = -1 * np.ones(labels.shape)
        labels_[clusters.core_sample_indices_] = labels[clusters.core_sample_indices_]
        labels = labels_

    # default fill of NA to catch all the points that are not active
    # we can't use np.nan yet because that breaks np.unique
    regions = np.zeros(shape=len(adata), dtype=np.int_)

    # relabel any points whose cluster is not large enough to 0 (noise)
    if min_size is not None:
        vals, counts = np.unique(labels, return_counts=True)
        vals_under_size = vals[counts < min_size]
        for v in vals_under_size:
            labels[labels == v] = -1

    counter = 1
    for c in np.unique(labels):
        size = np.count_nonzero(labels == c)
        this_inds = inds[labels == c]
        if c == -1 or size < min_size:
            # it just stays as is so do nothing
            continue
        else:
            regions[this_inds] = counter
            counter += 1

    n_regions = np.max(regions)
    if n_regions > 0:
        regions = pd.Categorical(regions,
                                 categories=np.arange(1, n_regions + 1)).remove_unused_categories()
        if output_key is not None:
            adata.obs[output_key] = regions
    else:
        regions = None

    if return_regions:
        return regions
    else:
        return n_regions


def hotspot_marker_genes(adata, eps, verbose=False, alpha=0.01, exclude_interaction=True,
                         neighborhood=True, interactions=None):
    """
    For each hotspot, identify genes that are differentially expressed against baselines of:
        1. other spatially nearby (but inactive cells)
        2. other cells of same interaction
    """

    filtered_interactions = adata.uns["interactions"]
    A = get_neighbor_adjacency(adata.obsm['spatial'], eps=eps)

    out = {v: {} for v in filtered_interactions["interaction_name"]}

    cutoff = -np.log10(alpha)

    for _, row in filtered_interactions.iterrows():
        interaction = row["interaction_name"]
        if interactions is not None and interaction not in interactions:
            continue

        ligand, receptor = row["ligand"], row["receptor"]

        # Go over all hotspots for this interaction
        try:
            c = adata.obs["hotspots_%s" % interaction]
        except KeyError:
            continue
        active_array = c.notnull()
        for v in np.unique(c):
            if np.isnan(v):
                continue
            hotspot_array = adata.obs[f"hotspots_{interaction}"] == v
            neighbor_array = np.logical_and(A.dot(np.array(hotspot_array, dtype=np.float64)) > 0,
                                            np.logical_not(hotspot_array))
            hotspot_inds = np.where(hotspot_array)[0]
            neighbor_inds = np.where(neighbor_array)[0]
            other_active_inds = \
                np.where(np.logical_and(active_array, np.logical_not(hotspot_array)))[0]

            # test differential expression against both
            # comparing hotspot to all active
            if np.min([len(hotspot_inds), len(other_active_inds), len(neighbor_inds)]) < 10:
                # very few cells - skip
                continue

            reg1 = differential_expression(adata, hotspot_inds, other_active_inds)
            significant = reg1["-log10(p)"] > cutoff
            reg2 = differential_expression(adata, hotspot_inds, neighbor_inds, alpha=alpha)

            # comparing hotspot to neighboring non-active
            if neighborhood:
                # identify genes that passed both significance tests
                significant = np.logical_and(significant, reg2["-log10(p)"] > cutoff)
                # check that they are differentially expressed the same way
                same_sign = np.sign(reg1["log2(fc)"]) == np.sign(reg2["log2(fc)"])

                selected = np.logical_and(significant, same_sign)
            else:
                selected = significant

            # filter out the interaction genes themsleves
            if exclude_interaction:
                selected = np.logical_and(selected,
                                          np.logical_not(reg1.index.isin([ligand, receptor])))

            # return a list of genes, sorted by magnitude of fold change
            genes = reg1.index[selected]
            avg_fc = (reg1["log2(fc)"] + reg2["log2(fc)"]) / 2
            avg_fc = avg_fc[selected]
            inds = np.argsort(-np.abs(avg_fc))
            genes = genes[inds]
            avg_fc = avg_fc[inds]

            out[interaction][v] = list(zip(genes, avg_fc))

    return out


def differential_expression(adata, inds_a, inds_b, alpha=0.001, max_fc=10, use_raw=True):
    if len(inds_a) == 0 or len(inds_b) == 0:
        raise ValueError("Indices list must be non-empty")

    if adata.raw is not None and use_raw:
        adata_source = adata.raw
    else:
        adata_source = adata
    array_a = adata_source[inds_a, :].X.toarray()
    array_b = adata_source[inds_b, :].X.toarray()

    # compute p-values
    U, pvalue = mannwhitneyu(array_a, array_b)
    _, pvalue = fdrcorrection(pvalue, alpha=alpha)
    pvalue = -np.log10(pvalue)

    # compute fold change
    x_med = np.mean(array_a, axis=0)
    y_med = np.mean(array_b, axis=0)
    # coupled with max_fc avoids division by 0
    eps = 1e-3
    x_med[x_med <= eps] = eps
    y_med[y_med <= eps] = eps
    fc = np.log2(x_med / y_med)
    fc[fc > max_fc] = max_fc
    fc[fc < -max_fc] = -max_fc

    return pd.DataFrame({'log2(fc)': fc, '-log10(p)': pvalue}, index=adata_source.var_names)


def gene_regulation(adata):
    res = {}
    filtered_interactions = adata.uns["interactions"]
    cutoffs = adata.uns["activity_significance_cutoff"]
    for i, row in filtered_interactions.iterrows():
        L, R = row["ligand"], row["receptor"]
        data = adata[:, R].X.toarray()
        expressed_inds = (data > np.quantile(data, 0.9)).reshape(-1)
        interaction_name = row["interaction_name"]
        print(interaction_name)
        try:
            v = np.array(adata.obs[f"activity_{interaction_name}"]) > cutoffs[interaction_name]
        except KeyError:
            # possible that an interaction has active cells but no hotspots so it doesn't get a
            # key here
            continue
        active_inds = np.logical_and(expressed_inds, v)
        inactive_inds = np.logical_and(expressed_inds, np.logical_not(v))

        res[interaction_name] = differential_expression(adata, active_inds, inactive_inds)

    return res


def get_all_hotspot_keys(adata):
    return [v for v in adata.obs if "hotspots_" in v]


def smooth(adata, alpha, k=6, threshold=None):
    smoothing_matrix = normalize(get_neighbor_adjacency(adata.obsm['spatial'], k=k), norm='l1',
                                 axis=1)
    if threshold is not None:
        smoothing_matrix = smoothing_matrix.multiply(get_neighbor_adjacency(adata.obsm['spatial'],
                                                                            eps=threshold))
    data = adata.X
    data = (1 - alpha) * data + alpha * (smoothing_matrix @ data)
    adata.X = data


def _compute_cutoff(gene_expression, log=False):
    if log:
        transformed_expr = np.exp(gene_expression) - 1
        cutoff = np.log(1 + threshold_otsu(transformed_expr))
    else:
        cutoff = threshold_otsu(gene_expression)
    return cutoff


def compute_hotspot_boundary(adata, hotspot_key, region, alpha_max=None, alpha_min=0.001):
    if alpha_max is None:
        alpha_max = 10
    v = adata.obs[hotspot_key]

    coords = adata.obsm['spatial']
    boundary = try_alpha_shape(coords[v == region, :], alpha_max=alpha_max, alpha_min=alpha_min)
    return boundary


def alpha_shape(coords, alpha):
    """
    https://gist.github.com/dwyerk/10561690#gistcomment-2819818

    Compute the alpha shape (concave hull) of a set
    of points.
    @param coords: Iterable container of points.
    @param alpha: alpha value to influence the
        gooeyness of the border. Smaller numbers
        don't fall inward as much as larger numbers.
        Too large, and you lose everything!
    """
    if len(coords) < 4:
        raise ValueError

    tri = Delaunay(coords)
    triangles = coords[tri.vertices]
    a = ((triangles[:, 0, 0] - triangles[:, 1, 0]) ** 2 + (
            triangles[:, 0, 1] - triangles[:, 1, 1]) ** 2) ** 0.5
    b = ((triangles[:, 1, 0] - triangles[:, 2, 0]) ** 2 + (
            triangles[:, 1, 1] - triangles[:, 2, 1]) ** 2) ** 0.5
    c = ((triangles[:, 2, 0] - triangles[:, 0, 0]) ** 2 + (
            triangles[:, 2, 1] - triangles[:, 0, 1]) ** 2) ** 0.5
    s = (a + b + c) / 2.0
    areas = (s * (s - a) * (s - b) * (s - c)) ** 0.5
    circums = a * b * c / (4.0 * areas)
    filtered = triangles[circums < (1.0 / alpha)]
    edge1 = filtered[:, (0, 1)]
    edge2 = filtered[:, (1, 2)]
    edge3 = filtered[:, (2, 0)]
    edge_points = np.unique(np.concatenate((edge1, edge2, edge3)), axis=0).tolist()
    m = geometry.MultiLineString(edge_points)
    triangles = list(polygonize(m))
    if len(triangles) == 0:
        #print('0')
        raise ValueError
    res = np.array(unary_union(triangles).boundary.coords)
    if len(res) == 4:
        #print('4')
        raise ValueError
    return res


def try_alpha_shape(coords, alpha_max, alpha_min):
    """
    Try to construct a shell around points for various alpha values, starting at alpha_max and
    decreasing by halves to alpha_min each time the construction fails.
    """
    if coords.shape[0] < 4:
        raise ValueError("coords must contain at least 4 points.")
    max_iters = 100
    alpha = alpha_max
    for _ in range(max_iters):
        try:
            boundary = alpha_shape(coords, alpha)
            return boundary
        except (NotImplementedError, ValueError):
            # This error is raised if the resulting geometry is not simply connected
            alpha = alpha * 0.5
            #print(alpha)
            if alpha < alpha_min:
                return None


def hotspot_closure(adata, alpha_max=10, edge_radius=0, verbose=False):
    """
    with Pool(processes) as p:
        keys = [v for v in adata.obs if "hotspots_" in v]
        points = adata.obsm['spatial']
        subcoords = []
        for key in tqdm(keys):
            for c in adata.obs[key].cat.categories:
                try:
                    boundary = compute_hotspot_boundary(adata, key, c, alpha_max=alpha_max)
                    path = mpl.path.Path(boundary)
                    adata.obs[key][path.contains_points(points, radius=-edge_radius)] = c
                except ValueError:
                    continue
    """
    keys = [v for v in adata.obs if "hotspots_" in v]
    points = adata.obsm['spatial']
    if verbose:
        keys = tqdm(keys)
    for key in keys:
        for c in adata.obs[key].cat.categories:
            try:
                boundary = compute_hotspot_boundary(adata, key, c, alpha_max=alpha_max)
                path = mpl.path.Path(boundary)
                adata.obs[key][path.contains_points(points, radius=-edge_radius)] = c
            except ValueError:
                continue


def hotspot_closure_sub(subcoords):
    try:
        return try_alpha_shape(subcoords, alpha_max=10,
                               alpha_min=0.001)
    except ValueError:
        return None



import numpy as np
import pandas as pd
import scanpy as sc

from tqdm import tqdm

from geospace.cci.spatial import *
from geospace.data import load_database


def compute_activity(adata, secreted_std, contact_threshold, permutations=100, K=0.5,
                     min_active_count=1,
                     sig_threshold=0.95, verbose=False, perform_permutation=False,
                     secreted_threshold_cutoff=2, save_activity=False,
                     interactions=None):
    """
    Compute the activity scores for all interactions across cells in adata. Note: filter_interactions must be run
    prior to this function. All output is stored in adata
    :param adata: anndata object
    :param permutations: number of permutation samples used in computing significance cutoffs (default 100)
    :param K: K parameter in Hill function
    :param min_active_count: minimum number of cells identified as active in order for interaction to be included in
    result (default 0)
    :param sig_threshold: quantile of permutation test bootstrap distribution used for significance cutoff
    :param secreted_std: Standard deviation of secreted Gaussian diffusion
    :param secreted_threshold_cutoff: Maximum number of standard deviations to compute
    :param contact_threshold: Maximum transport distance for contact interactions (must also be adjacent in Delauney
    triangulation)
    :param verbose: If true, prints out additional information while running (default: false)
    :param save_activity: If true, computed activity scores are saved in adata.obs
    """

    # Check that filter_interactions has been run
    try:
        filtered_interactions = adata.uns["interactions"]
    except KeyError:
        interaction = load_database()
        filter_interactions(interaction, adata)
        filtered_interactions = adata.uns["interactions"]

    # Get a reduced view of adata containing only relevant genes
    adata_filtered = filter_adata(adata)
    compute_spatial_transport_matrices(adata,
                                       secreted_std=secreted_std,
                                       secreted_threshold_cutoff=secreted_threshold_cutoff,
                                       contact_threshold=contact_threshold)
    transport_secreted = adata.obsp['transport_secreted']
    transport_contact = adata.obsp['transport_contact']

    # Here we use the raw values
    try:
        expr = adata_filtered.X.toarray()
    except AttributeError:
        expr = adata_filtered.X
    # expr = np.log(1+expr)
    # Normalize each column to 0-1
    expr /= np.mean(expr, axis=0)
    expr = scipy.sparse.csc_matrix(expr)
    # compute the L_i R_j score for each interaction and pair of genes
    activity_matrix_cols = {}
    pathways = {}
    interaction_to_pathway = {}

    if perform_permutation:
        permutation_vectors = []
        n_cells = len(adata_filtered)
        expr_boot = []
        # A_boot = []
        for i in range(permutations):
            permutation = np.random.permutation(n_cells)
            permutation_vectors.append(permutation)
            expr_boot.append(expr[permutation, :])

    # track which interactions still remain after activity filtering
    cutoffs = {}
    rows = filtered_interactions.iterrows()
    if verbose:
        print("Computing activity scores")
        rows = tqdm(rows, total=filtered_interactions.shape[0])
    for k, row in rows:
        pathway = row["pathway_name"]
        interaction_name = row["interaction_name"]

        if interactions is not None and interaction_name not in interactions:
            pass

        if row["secreted"]:
            transport = transport_secreted
        else:
            transport = transport_contact

        if pathway not in pathways:
            pathways[pathway] = 1
        else:
            pathways[pathway] += 1

        i = np.where(adata_filtered.var_names == row["ligand"])[0]
        j = np.where(adata_filtered.var_names == row["receptor"])[0]

        ligand_i, receptor_j = expr[:, i], expr[:, j]
        activity = receptor_j.multiply(transport.dot(ligand_i))
        # Apply hill function
        activity.data = activity.data / (K + activity.data)
        activity = activity.toarray().reshape(-1)

        # filter out interactions with a low number of possibly interacting cells
        nonzero = np.count_nonzero(activity)
        if nonzero < min_active_count:
            continue

        activity_matrix_cols[interaction_name] = activity

        if perform_permutation:
            vals = np.array([])
            for kk in range(permutations):
                ligand_i_boot, Rj_boot = expr_boot[kk][:, i], expr_boot[kk][:, j]

                activity_boot = Rj_boot.multiply(transport.dot(ligand_i_boot))
                activity_boot.data = activity_boot.data / (K + activity_boot.data)
                activity_boot = activity_boot.toarray().reshape(-1)
                vals = np.concatenate([vals, activity_boot])
            # TODO: check if we can avoid the zeros here to speed it up a little
            cutoff = np.quantile(vals, sig_threshold)
            cutoffs[interaction_name] = cutoff

        interaction_to_pathway[interaction_name] = pathway

    if len(activity_matrix_cols) == 0:
        warnings.warn("No significant interactions were found in the dataset")
        return

    activity_matrix = pd.DataFrame(activity_matrix_cols)

    # Filter out inactive interactions
    filtered_interactions = filtered_interactions[filtered_interactions["interaction_name"].isin(activity_matrix_cols)]
    adata.uns["interactions"] = filtered_interactions

    if perform_permutation:
        adata.uns["activity_significance_cutoff"] = cutoffs
    if save_activity:
        pd.concat([adata.obs, activity_matrix], axis=1)

    return activity_matrix


def filter_interactions(interaction, adata, filter_same=True, pathway_filter=None,
                        interaction_filter=None):
    """
    Given a particular adata object, filter out only the LR interactions in the database for which
    both the L and the R are present in the genes in the adata. Modifies the database accordingly.
    """
    st_genes = {v for v in adata.var_names}
    interaction_index = []

    lr_set = set()
    r_set = set()
    pathways = set()

    for i, row in interaction.iterrows():
        pathway = row["pathway_name"]
        if pathway_filter is not None and pathway not in pathway_filter:
            continue

        if interaction_filter is not None and row["interaction_name"] not in interaction_filter:
            continue

        ligand, receptor = row["ligand"], row["receptor"]

        if filter_same and ligand == receptor:
            continue

        if ligand in st_genes and receptor in st_genes:
            lr_set.add(row["interaction_name"])
            r_set.add(receptor)
            pathways.add(row["pathway_name"])
            interaction_index.append(i)

    interaction_index = np.array(interaction_index)
    filtered_interactions = interaction.iloc[interaction_index].reset_index(drop=True)

    adata.uns["interactions"] = filtered_interactions

    if len(filtered_interactions) == 0:
        warnings.warn("No interactions from database matched in dataset.")
        # TODO: remove the debug error
        raise ValueError

    # return the list of identified interactions
    return list(filtered_interactions["interaction_name"])


def filter_adata(adata):
    """Filter out genes that are not part of a known LR interaction (or for which the partner
    is not present)"""
    filtered_interactions = adata.uns["interactions"]
    ligand_genes = np.unique(filtered_interactions["ligand"])
    receptor_genes = np.unique(filtered_interactions["receptor"])
    ligand_receptor_genes = np.unique(np.concatenate((ligand_genes, receptor_genes)))
    gene_names = np.array([v for v in adata.var_names])
    ligand_receptor_gene_inds = np.array([np.where(gene_names == gene)[0]
                                          for gene in ligand_receptor_genes]).flatten()
    adata_filtered = adata[:, ligand_receptor_gene_inds]
    adata_filtered.var_names = gene_names[ligand_receptor_gene_inds]
    return adata_filtered


def compute_active_sub(M, K):
    receptor_active = np.array(np.sum(M, axis=0)).reshape(-1)
    receptor_active = receptor_active / (K + receptor_active)
    return receptor_active


def compute_active_matrix(adata, P, K=0.5, smoothing=0.0, A=None, min_active_count=0):
    # Compute cell-by-active matrix
    active_matrix = np.zeros((len(adata), len(P)), dtype=np.float64)
    columns = []
    col_inds = []
    for k, (interaction_name, M) in enumerate(P.items()):
        v = compute_active_sub(M, K)
        # filter out those below the precomputed significance level
        # +v[v < cutoffs[interaction_name]] = 0
        if np.count_nonzero(v) < min_active_count:
            continue
        columns.append(interaction_name)
        col_inds.append(k)
        active_matrix[:, k] = v
    active_matrix = active_matrix[:, np.array(col_inds)]

    active_matrix = pd.DataFrame(active_matrix, columns=columns)
    return active_matrix, columns


def combine_pathway(filtered_interactions, active_matrix, type="arithmetic"):
    """
    Take an activity matrix and combine all activity on a per-pathway basis
    """

    interactions_by_pathway = {k: [] for k in
                               np.unique(filtered_interactions["pathway_name"])}
    for _, row in filtered_interactions.iterrows():
        interactions_by_pathway[row["pathway_name"]].append(row["interaction_name"])
    res = {}
    if type == "geometric":
        mean = scipy.stats.mstats.gmean
    elif type == "arithmetic":
        mean = np.mean
    else:
        raise ValueError("Type must be either 'geometric' or 'arithmetic'")
    for pathway, interactions in interactions_by_pathway.items():
        try:
            v = mean(active_matrix.loc[:, interactions], axis=1)
        except KeyError as e:
            print(list(active_matrix))
            raise e
        res[pathway] = v
    active_matrix_pathway = pd.DataFrame.from_dict(res)

    return active_matrix_pathway


def combine_receptor(filtered_interactions, active_matrix, type="arithmetic"):
    """
    Take an activity matrix and combine all activity on a per-receptor basis
    """

    interactions_by_receptor = {k: [] for k in
                                np.unique(filtered_interactions["receptor"])}
    for _, row in filtered_interactions.iterrows():
        interactions_by_receptor[row["receptor"]].append(row["interaction_name"])

    res = {}
    if type == "geometric":
        mean = scipy.stats.mstats.gmean
    elif type == "arithmetic":
        mean = np.mean
    else:
        raise ValueError("Type must be either 'geometric' or 'arithmetic'")
    for pathway, interactions in interactions_by_receptor.items():
        res[pathway] = mean(active_matrix.loc[:, interactions], axis=1)
    active_matrix_receptor = pd.DataFrame.from_dict(res)

    return active_matrix_receptor


def cell_type_activity(adata, format, type_label="class"):
    """
    Compute the average activity over all
    :param adata: anndata object after running compute_activity
    :param format:
    :param type_label: categorical label in adata.obs that defines cell types (or other exclusive grouping)
    :return:
    """

    # Either process for LR pairs, receptors, or pathways
    if format != "LR":
        raise NotImplementedError

    cols = list(adata.uns["interactions"]["interaction_name"])

    cols = ["activity_%s" % k for k in cols]
    # print(cols)
    active_matrix = adata.obs.loc[:, cols]

    types = np.unique(adata.obs[type_label])
    rows = []
    significant_mat = np.zeros((len(types), active_matrix.shape[1]))

    active_matrix_array = np.array(active_matrix)
    for k, t in enumerate(types):
        # test for significance
        active_matrix_sub = active_matrix_array[(adata.obs[type_label] == t)]
        sub_activity = np.mean(np.array(active_matrix)[(adata.obs[type_label] == t)], axis=0)

        U, p = scipy.stats.mannwhitneyu(active_matrix_sub, active_matrix_array,
                                        alternative='greater')
        significant = p < 0.05
        rows.append(sub_activity)
        significant_mat[k, :] = significant

    # take out the "activity_"
    cols = [k[9:] for k in active_matrix.columns]
    out = pd.DataFrame(rows, columns=cols, index=types)

    def style_significant(v):
        return np.where(significant_mat, "font-weight: bold", "font-weight: lighter")

    style = out.style.apply(style_significant, axis=None)
    return out, style


import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
import matplotlib.colors as colors
import pandas as pd
import numpy as np
import seaborn as sns
from geospace.plot.plot import spatial, scale


def hotspots(adata, hotspot_key, ax=None, show=True, correct_categorical=False, labels=True,
             **kwargs):
    if ax is None:
        ax = plt.gca()
    if "title" not in kwargs:
        kwargs["title"] = " - ".join([v.capitalize() for v in hotspot_key.split("_")])

    key = "hotspots_%s" % hotspot_key

    if correct_categorical and not hasattr(adata.obs[key], 'cat'):
        v = pd.Categorical(adata.obs[key])
        v.categories = v.categories.astype(np.int_)
        adata.obs[key] = v

    if labels:
        legend_loc = "on data"
    else:
        legend_loc = None
    spatial(adata, color=key,
            ax=ax, show=False, legend_loc=legend_loc, na_in_legend=False,
            **kwargs)

    if show:
        plt.show()


def hotspots_multiple(adata, nrows, ncols, hotspot_keys=None, figsize=None):
    # interactions = ["WNT5A_FZD2", "WNT5B_FZD2", "WNT11_FZD2", "APLN_APLNR",
    #                "FGF5_FGFR1", "FGF10_FGFR1", "FGF15_FGFR1", "FGF5_FGFR2"]

    if hotspot_keys is None:
        obs_keys = list(adata.obs)
        hotspot_keys = [v[9:] for v in obs_keys if "hotspots_" in v]
        nrows = int(np.ceil(len(hotspot_keys) / ncols))

    fig, axs = plt.subplots(nrows, ncols, figsize=figsize)

    for i, hotspot_key in enumerate(hotspot_keys):
        cur_row = i // ncols
        cur_col = i % ncols
        hotspots(adata, hotspot_key, ax=axs[cur_row, cur_col], show=False)

    for i in range(nrows):
        for j in range(ncols):
            axs[i, j].set_xlabel('')
            axs[i, j].set_ylabel('')
    plt.tight_layout()


def multi_hotspots(adata, show=True, inds=None, spotcolor=None, ax=None, num=None, linewidth=0.5,
                   legend_ncol=None, alpha_img=1.0, color_type=None, alpha=150, base_color=None,
                   frameon=False,
                   **kwargs):
    if color_type is None:
        color_type = "index"

    cm1 = mpl.cm.get_cmap('tab20')
    legend_handles = []
    legend_labels = []
    create_legend = True

    if color_type == "genes":
        create_legend = False
        vals = [len(v) for k, v in adata.uns["multi_hotspots"].items()]
        min_genes = min(vals)
        max_genes = max(vals)
        norm = colors.LogNorm(vmin=min_genes, vmax=max_genes)
        cmap = mpl.cm.YlOrRd

        sm = mpl.cm.ScalarMappable(norm=norm, cmap=cmap)


    try:
        scale_factor = list(adata.uns['spatial'].values())[0]['scalefactors']['tissue_hires_scalef']
    except KeyError:
        if scale is not None:
            scale_factor = scale
        else:
            scale_factor = 1

    if ax is None:
        fig, ax = plt.subplots()

    spatial(adata, color=base_color, ax=ax, alpha_img=alpha_img, frameon=frameon,
            na_color=spotcolor,
            show=False, **kwargs)

    hotspot_keys = [v for v in adata.obs if "hotspots_multi" in v]
    if inds is None:
        # Draw boundaries around all considered regions
        inds = list(range(len(hotspot_keys)))
    for c, idx in enumerate(inds):
        if num is not None and idx >= num:
            break
        if color_type == "index":
            color = cm1(c)
        elif color_type == "genes":
            color = sm.to_rgba(len(adata.uns["multi_hotspots"][str(idx)]))
        else:
            # assume it is a vector with colors
            color = colors.to_rgba(color_type[idx])

        color_int = [int(255 * v) for v in color]
        color_string = "#%02x%02x%02x%02x" % (color_int[0], color_int[1], color_int[2], alpha)
        color_string_noalpha = "#%02x%02x%02x" % (color_int[0], color_int[1], color_int[2])

        if True or idx < 21:
            legend_labels.append(idx)
            #legend_handles.append(Patch(facecolor=color_string, edgecolor='k',
            #                            label=idx))
            legend_handles.append(Line2D([0], [0], marker='o', color='w', label=idx,
                   markerfacecolor=color_string_noalpha, markersize=8))


        try:
            boundary = scale_factor*adata.uns["multi_boundaries"][str(idx)]
        except KeyError:
            raise ValueError("Need to compute boundaries first")
        ax.fill(boundary[:, 0], boundary[:, 1], color_string)
        ax.plot(boundary[:, 0], boundary[:, 1], linewidth=linewidth, color="black")
    if legend_ncol is not None and create_legend:
        ax.legend(legend_handles, legend_labels, loc="center left", bbox_to_anchor=(1, 0.5),
                  ncol=legend_ncol, frameon=False)
        #ax.legend(legend_handles, legend_labels, loc="upper center", bbox_to_anchor=(0.5, 0.0),
        #          ncol=legend_ncol)
    elif color_type == "genes":
        ax.get_figure().colorbar(sm, orientation='vertical', label='coexpressed genes',
                                 shrink=0.8)


    #plt.tight_layout()
    if show:
        plt.show()





# TODO: put this file somewhere else

import scipy.stats

import numpy as np

from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import adjusted_rand_score

from geospace.hotspot.hotspot import _compute_cutoff


def entropy(adata, labels, pca=True, dim=16, depth=10, output_label="entropy", use_raw=False):
    X, y = get_data_and_labels(adata, labels, pca, dim, use_raw=use_raw)
    clf = RandomForestClassifier(max_depth=depth)
    clf.fit(X, y)

    class_probs = clf.predict_proba(X)
    entropy = scipy.stats.entropy(class_probs, axis=1)

    adata.obs[output_label] = entropy
    entropy_binary = np.zeros(entropy.shape)
    cutoff = _compute_cutoff(entropy)
    entropy_binary[entropy > cutoff] = 1
    # TODO: figure out all these labels
    adata.obs["entropy_binary"] = entropy_binary

    # TODO: also compute the scaled entropy or whatever
    scaled_entropy = np.empty(entropy.shape)
    for c in y:
        scaled_entropy[y == c] = scipy.stats.zscore(entropy[y == c])
    adata.obs["scaled_entropy"] = scaled_entropy


def forest_depth_analysis(adata, labels, depths=None, pca=True, dim=16, verbose=True):
    # TODO: rewrite this with proper cross validation and stuff
    if depths is None:
        depths = list(range(2, 15))
    X, y = get_data_and_labels(adata, labels, pca, dim)
    # TODO: shuffle properly
    train_cutoff = round(0.8 * len(adata))
    X_train = X[:train_cutoff, :]
    y_train = y[:train_cutoff]

    X_test = X[train_cutoff:, :]
    y_test = y[train_cutoff:]

    train_scores = []
    test_scores = []
    test_aris = []
    for depth in depths:
        clf = RandomForestClassifier(max_depth=depth)
        clf.fit(X_train, y_train)
        train_score = clf.score(X_train, y_train)
        test_score = clf.score(X_test, y_test)
        ari = adjusted_rand_score(clf.predict(X_test), y_test)
        train_scores.append(train_score)
        test_scores.append(test_score)
        test_aris.append(ari)
        if verbose:
            print("Depth %d: Train %f - Test %f - ari %f" % (depth, train_score, test_score, ari))

    return depths, test_aris


def get_data_and_labels(adata, labels, pca, dim, use_raw=False):
    if use_raw:
        adata_to_use = adata.raw
    else:
        adata_to_use = adata
    try:
        data = adata_to_use.X.toarray()
    except AttributeError:
        data = adata_to_use.X
    if pca:
        # Reduce dimensions using PCA
        pca_transform = PCA(n_components=dim)
        X = pca_transform.fit_transform(data)
    else:
        X = data

    # labels provided either as list or key in adata.obs
    try:
        y = adata.obs[labels]
    except IndexError:
        y = labels

    return X, y


import pandas as pd
import numpy as np
from itertools import combinations, chain
from collections import Counter
import matplotlib.path
import time
from multiprocessing import Pool

from scipy.sparse import csr_matrix, coo_matrix

import scanpy as sc
from tqdm import tqdm
import networkx as nx

from .hotspot import get_all_hotspot_keys, compute_hotspot_boundary


def coexpression_hotspots(adata, jaccard_matrix=None, hotspot_lookup=None, threshold=0.7,
                          min_size=0, min_genes=5, cutoff=None, verbose=False,
                          processes=1, divisions=10):
    if jaccard_matrix is None:
        if verbose:
            print("Computing hotspot similarity matrix.")
        jaccard_matrix, hotspot_lookup = hotspot_jaccard(adata, processes=processes,
                                                         verbose=verbose, divisions=divisions)

    multigene_hotspots = hotspot_groups(jaccard_matrix, hotspot_lookup,
                                        threshold=threshold, min_genes=min_genes)

    median_sizes = []
    min_size = 5
    #cutoff = 0.5
    for idx, multigene_hotspot in enumerate(multigene_hotspots):
        sizes = []
        for key, v in multigene_hotspot:
            hotspot_size = np.count_nonzero(adata.obs[key] == v)
            sizes.append(hotspot_size)
        median_sizes.append(int(np.median(sizes)))
    multi_arrays = {}
    new_multigene_hotspots = []
    score_dict = {}
    for idx, (multigene_hotspot, median_size) in enumerate(
            sorted(zip(multigene_hotspots, median_sizes), key=lambda x: -x[1])):
        if median_size < min_size:
            break
        hotspots_per_spot = np.zeros(adata.shape[0])

        for idx2, (key, v) in enumerate(multigene_hotspot):
            hotspots_per_spot += (adata.obs[key] == v).astype(np.int_)

        if cutoff is None:
            inds = np.argsort(hotspots_per_spot)[::-1][:median_size]
        else:
            inds = np.where(hotspots_per_spot > cutoff * len(multigene_hotspot))[0]

        score = np.sum(hotspots_per_spot[inds]) / np.sum(hotspots_per_spot)
        score_dict[idx] = score
        # compute essence of overlap

        v = np.full(adata.shape[0], np.nan)
        v[inds] = 0
        multi_arrays[idx] = pd.Categorical(v)
        new_multigene_hotspots.append(multigene_hotspot)

    multi_arrays = {f"hotspots_multi_{idx}": v for idx, (k, v) in enumerate(multi_arrays.items())}
    # adata.obs = pd.concat([adata.obs, pd.DataFrame(multi_arrays, index=adata.obs.index)], axis=1)
    tmp_df = adata.obs.merge(pd.DataFrame(multi_arrays, index=adata.obs.index), how="outer",
                             left_index=True,
                             right_index=True, suffixes=('_old', None))
    tmp_df.drop(tmp_df.filter(regex='_old$').columns.tolist(), axis=1, inplace=True)
    adata.obs = tmp_df
    new_multigene_hotspots = {idx: v for idx, v in enumerate(new_multigene_hotspots)}
    adata.uns["multi_hotspots"] = new_multigene_hotspots

    return len(multi_arrays)


def multi_hotspot_groupings(adata, num_groupings=None):
    multi_hotspots = [v for v in adata.obs if "hotspots_multi" in v]
    multi_hotspots_df = sc.get.obs_df(adata, keys=multi_hotspots)
    # compute to matrix, 0 -> 1, nan -> 0
    multi_hotspots_array = np.nan_to_num(np.array(multi_hotspots_df, dtype=np.float32) + 1).astype(
        np.int_)
    type_counter = Counter()
    for idx in range(adata.shape[0]):
        t = np.where(multi_hotspots_array[idx, :])
        type_counter[t] += 1

    return sorted(type_counter.items(), key=lambda x: -x[1])


def hotspot_decomposition(adata, compute_type_counts=False, neg_weight=1.0):
    # assumes that single-gene hotspots and multi-gene hotspots have already been computed
    all_hotspots = [v for v in adata.obs if "hotspots_" in v and "multi" not in v]
    multi_hotspots = [v for v in adata.obs if "hotspots_multi" in v]

    multi_hotspots_df = sc.get.obs_df(adata, keys=multi_hotspots)
    # compute to matrix, 0 -> 1, nan -> 0
    multi_hotspots_array = np.nan_to_num(np.array(multi_hotspots_df, dtype=np.float32) + 1).astype(
        np.int_)

    weights_dict = hotspot_decomposition_keys(adata, all_hotspots, multi_hotspots_array,
                                              neg_weight=neg_weight)

    type_counter = Counter()
    if compute_type_counts:
        types = [tuple(np.where(weights)[0]) for (weights, _) in weights_dict.values() if
                 np.sum(weights) > 0]
        for t in types:
            for subset in powerset(t):
                type_counter[subset] += 1
        return weights_dict, sorted(type_counter.items(), key=lambda x: -x[1])

    return weights_dict


def hotspot_decomposition_keys(adata, keys, multi_hotspots_array, neg_weight=None):
    weights_dict = {}
    for hotspot in tqdm(keys):
        hotspot_vector = np.array(adata.obs[hotspot], dtype=np.float32)
        hotspot_vector[pd.notnull(hotspot_vector)] = 1
        hotspot_vector = np.nan_to_num(np.array(hotspot_vector, dtype=np.float32)).astype(np.int_)

        weights, score = hotspot_decomposition_sub(hotspot_vector, multi_hotspots_array,
                                                   neg_weight=neg_weight)
        hotspot_name = hotspot[9:]
        weights_dict[hotspot_name] = (weights, score)

    return weights_dict


def hotspot_decomposition_sub(hotspot_vector, multi_hotspots_array, neg_weight=None):
    if neg_weight is None:
        neg_weight = 1.0
    original_size = np.sum(hotspot_vector.flatten()).astype(np.float64)
    weights = np.zeros(multi_hotspots_array.shape[1])
    score = 0.0

    while True:
        overlap = np.sum(np.multiply(hotspot_vector.reshape(-1, 1), multi_hotspots_array), axis=0)
        neg_overlap = np.sum(np.multiply(1 - hotspot_vector.reshape(-1, 1), multi_hotspots_array),
                             axis=0)
        score_change = overlap - neg_weight * neg_overlap
        max_ind = np.argmax(score_change)
        if score_change[max_ind] > 0:
            weights[max_ind] = 1
            intersects = np.logical_and(hotspot_vector, multi_hotspots_array[:, max_ind])
            hotspot_vector[intersects] = 0
            score += score_change[max_ind]
        else:
            break

    score /= original_size
    # check if we met the minimum coverage
    # if np.sum(hotspot_vector.flatten())/original_size < min_covered_frac:
    # we didn't so return nothing
    #    return np.zeros(multi_hotspots_array.shape[1]), 0.0
    return weights, score


def hotspot_jaccard(adata, keys=None, processes=4, verbose=False, divisions=5):
    if keys is None:
        keys = get_all_hotspot_keys(adata)

    num_hotspots = 0
    mean_size = 0
    indptr_counter = 0
    offsets = {}

    indices_list = []
    indptr = [0]

    hotspot_sizes = []

    hotspot_lookup = []

    for idx, hotspot_key in enumerate(keys):
        offsets[hotspot_key] = num_hotspots
        v = pd.Categorical(adata.obs[hotspot_key])
        n_hotspots = len(v.categories)
        for c in range(n_hotspots):
            inds = np.where(v.codes == c)[0]
            indices_list.append(inds)
            indptr_counter += len(inds)
            indptr.append(indptr_counter)
            hotspot_sizes.append(len(inds))
            mean_size += len(inds)
            hotspot_lookup.append((hotspot_key, v.categories[c]))

        num_hotspots += n_hotspots

    # print(f"{counter} total hotspots.")

    indices = np.concatenate(indices_list)
    data = np.ones(indices.shape)
    hotspot_matrix = csr_matrix((data, indices, indptr)).tocsc()

    hotspot_sizes = np.array(hotspot_sizes, dtype=np.float32)

    # Iterate over columns, each representing a spot and the hotspots it belongs to
    num_pairs = 0
    combination_iterators = []
    num_pairs_arr = []
    for col in range(hotspot_matrix.shape[1]):
        hotspots = hotspot_matrix[:, col].nonzero()[0]
        l = len(hotspots)
        num_pairs += l * (l - 1) // 2
        num_pairs_arr.append(l * (l - 1) // 2)
        combination_iterators.append(combinations(hotspots, 2))
    # print(f"{num_pairs} total overlap pairs.")

    # TODO: maybe handle the case we can't import multiprocessing (does that happen?)
    if processes is not None:
        p = Pool(processes)
    # Spatially divide the spots
    coords = adata.obsm['spatial']
    xmin, ymin = np.min(coords, axis=0)
    xmax, ymax = np.max(coords, axis=0)
    xmax += 0.00001
    ymax += 0.00001
    xdiff = xmax - xmin
    ydiff = ymax - ymin
    # Partition up all the spots based on a grid
    data_list = []
    for idx in range(divisions):
        for idy in range(divisions):
            cur_xmin = xmin + (idx / divisions) * xdiff
            cur_xmax = xmin + ((idx + 1) / divisions) * xdiff

            cur_ymin = ymin + (idy / divisions) * ydiff
            cur_ymax = ymin + ((idy + 1) / divisions) * ydiff
            in_current_element = np.all([coords[:, 0] >= cur_xmin, coords[:, 0] < cur_xmax,
                                         coords[:, 1] >= cur_ymin, coords[:, 1] < cur_ymax],
                                        axis=0)
            inds = np.where(in_current_element)[0]
            if len(inds) == 0:
                continue
            cur_iters = []
            cur_num_pairs = 0
            for ind in inds:
                cur_iters.append(combination_iterators[ind])
                cur_num_pairs += num_pairs_arr[ind]
            data_list.append((cur_iters, cur_num_pairs, num_hotspots))

    num_tasks = len(data_list)

    overlap_matrix = csr_matrix((num_hotspots, num_hotspots))

    if verbose:
        list_wrapper = tqdm
    else:
        list_wrapper = lambda x, total: x

    if processes is not None:
        for sub_matrix in list_wrapper(p.imap_unordered(_multi_wrapper, data_list),
                                       total=num_tasks):
            overlap_matrix += sub_matrix
            del sub_matrix
    else:
        for sub_matrix in list_wrapper((_multi_wrapper(data) for data in data_list),
                                       total=num_tasks):
            overlap_matrix += sub_matrix
            del sub_matrix

    if processes is not None:
        p.close()

    del hotspot_matrix
    overlap_matrix = overlap_matrix.tocoo()

    # Compute the jaccard matrix
    if verbose:
        print("Computing jaccard matrix")
    row_hotspot_size = hotspot_sizes[overlap_matrix.row]
    col_hotspot_size = hotspot_sizes[overlap_matrix.col]
    overlap_count = overlap_matrix.data
    jaccard_matrix_data = row_hotspot_size + col_hotspot_size
    jaccard_matrix_data -= overlap_count
    jaccard_matrix_data = overlap_count / jaccard_matrix_data
    if verbose:
        print("Converting jaccard matrix")
    jaccard_matrix = coo_matrix(
        (jaccard_matrix_data, (overlap_matrix.row, overlap_matrix.col)),
        overlap_matrix.shape).tocsr()

    return jaccard_matrix, hotspot_lookup


def _multi_wrapper(data):
    iters, num_pairs, num_hotspots = data
    return compute_overlap_matrix(iters, num_pairs, num_hotspots, verbose=False)


def compute_overlap_matrix(iterator_list, num_pairs, num_hotspots, verbose=True):
    t0 = time.time()
    idx = 0
    if verbose:
        print(f"Constructing overlap matrix with {num_pairs} pairs.")
        #print(f"Current memory usage: {psutil.Process().memory_info().rss/(1024*1024)}")
    row = np.empty(num_pairs, dtype=np.int32)
    col = np.empty(num_pairs, dtype=np.int32)
    if verbose:
        iterator_list = tqdm(iterator_list)
    for it in iterator_list:
        for i, j in it:
            row[idx], col[idx] = i, j
            idx += 1
    data = np.ones(num_pairs, dtype=np.float32)
    if verbose:
        print("Converting overlap matrix")
    overlap_matrix = coo_matrix((data, (row, col)), (num_hotspots, num_hotspots),
                                             dtype=np.float32).tocsr()
    del data, row, col
    if verbose:
        t1 = time.time()
        print(f"Completed in {t1 - t0} seconds.")
    return overlap_matrix


def hotspot_groups(jaccard_matrix, hotspot_lookup, threshold, min_genes=5):
    jaccard_matrix_coo = jaccard_matrix.tocoo()
    above_threshold = jaccard_matrix_coo.data > threshold
    row = jaccard_matrix_coo.row[above_threshold]
    col = jaccard_matrix_coo.col[above_threshold]
    data = jaccard_matrix_coo.data[above_threshold]
    jaccard_matrix_coo.data = data
    jaccard_matrix_coo.row = row
    jaccard_matrix_coo.col = col
    jaccard_matrix_reduced = jaccard_matrix_coo.tocsr()
    del jaccard_matrix_coo

    g = nx.from_scipy_sparse_matrix(jaccard_matrix_reduced)
    g2 = nx.k_core(g, k=min_genes)

    multigene_hotspots = []
    for c in nx.connected_components(g2):
        multigene_hotspot = [hotspot_lookup[i] for i in c]
        multigene_hotspots.append(multigene_hotspot)

    return multigene_hotspots


def compute_multi_boundaries(adata, alpha_max, alpha_min):
    multi_hotspots = [v for v in adata.obs if "hotspots_multi" in v]
    adata.uns["multi_boundaries"] = {}
    for idx, hotspot_key in tqdm(enumerate(multi_hotspots), total=len(multi_hotspots)):
        boundary = compute_hotspot_boundary(adata, hotspot_key, 0, alpha_max=alpha_max,
                                            alpha_min=alpha_min)
        # saving of anndata converts dictionary indices in .uns to strings anyway
        adata.uns["multi_boundaries"][str(idx)] = boundary


def multi_closure(adata):
    hotspot_keys = [v for v in adata.obs if "hotspots_multi" in v]
    points = adata.obsm['spatial']
    for idx, hotspot_key in enumerate(hotspot_keys):
        boundary = adata.uns["multi_boundaries"][idx]
        path = matplotlib.path.Path(boundary)
        v = pd.Categorical(path.contains_points(points, radius=-180), categories=(True,))
        v.categories = [0]
        adata.obs[hotspot_key] = v


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(1, len(s) + 1))



import pandas as pd
import scanpy as sc
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import seaborn as sns


def tracks_plot(adata, marker_genes, width=12, track_height=0.25,
                fontsize=None, use_raw=False, exp_transform=False,
                other_genes=None, palette=None, gene_name_italic=False):
    if other_genes is None:
        other_genes = []
    if gene_name_italic:
        italicstyle = 'italic'
    else:
        italicstyle = None
    keys = []
    tracksplot_genes = []
    categories = []
    category_labels = []
    label_rotation = 0
    for k, genes in marker_genes.items():
        categories.append(k)
        category_labels.append(f"CH{k}")

        keys.append(f'hotspots_multi_{k}')

        for gene in genes:
            if gene not in keys:
                keys.append(gene)
                tracksplot_genes.append(gene)

    for gene in other_genes:
        if gene not in keys:
            keys.append(gene)
            tracksplot_genes.append(gene)


    # Prepare the data for the tracksplot
    obs_tidy = sc.get.obs_df(adata, keys=keys, use_raw=use_raw)

    # Create the tracksplot
    nbins = 10

    dendro_height = 0

    groupby_height = 0.24
    # +2 because of dendrogram on top and categories at bottom
    num_rows = len(tracksplot_genes)
    #width = 12
    #track_height = 0.25

    height_ratios = [track_height] * len(tracksplot_genes)
    height = 2 * sum(height_ratios)

    # TODO: sort out colors
    groupby_colors = None
    if palette is not None:
        groupby_colors = sns.color_palette(palette)[:len(categories)]

    if groupby_colors is None and len(categories) <= 10:
        groupby_colors = sns.color_palette("deep")[:len(categories)]

    fig = plt.figure(figsize=(width, height))
    axs = gridspec.GridSpec(
        ncols=2,
        nrows=num_rows,
        wspace=1.0 / width,
        hspace=0,
        height_ratios=height_ratios,
        width_ratios=[width, 0.14],
    )
    axs_list = []
    first_ax = None
    for idx, gene in enumerate(tracksplot_genes):
        ax_idx = idx
        if first_ax is None:
            ax = fig.add_subplot(axs[ax_idx, 0])
            first_ax = ax
        else:
            ax = fig.add_subplot(axs[ax_idx, 0], sharex=first_ax)
        axs_list.append(ax)
        for cat_idx, category in enumerate(categories):
            #obs_cur_cat = obs_tidy[obs_tidy["hotspots_%s"%interaction] == category]
            obs_cur_cat = obs_tidy[pd.notnull(obs_tidy[f'hotspots_multi_{category}'])]
            expression_values = np.sort(
                obs_cur_cat.loc[:, gene].to_numpy())  # Get the expression_values
            if exp_transform:
                expression_values = np.exp(expression_values)-1
            average_expressions = np.zeros(nbins)

            num = int(np.floor(np.size(expression_values) / nbins))

            for ave_idx in range(nbins):
                if ave_idx < nbins - 1:
                    try:
                        average_expressions[ave_idx] = np.mean(
                            expression_values[num * (ave_idx):num * (1 + ave_idx)])
                    except ZeroDivisionError:
                        # TODO: this probably isn't necessary
                        print(expression_values.shape)
                        print(num * (ave_idx), num * (1 + ave_idx))
                else:
                    average_expressions[ave_idx] = np.mean(expression_values[num * (ave_idx):])

            if groupby_colors is not None:
                cc = groupby_colors[cat_idx]
            else:
                cc = None
            ax.fill_between(
                range(cat_idx * nbins, (cat_idx + 1) * nbins),
                0,
                average_expressions,
                lw=0.1,
                color=cc,
            )

        # remove the xticks labels except for the last processed plot.
        # Because the plots share the x axis it is redundant and less compact
        # to plot the axis for each plot
        if idx < len(tracksplot_genes) - 1:
            try:
                ax.tick_params(labelbottom=False, labeltop=False, bottom=False, top=False,
                               size=fontsize)
                ax.set_xlabel('')
            except TypeError:
                # this fails safely in jupyter notebooks
                pass
        ax.spines['left'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.grid(False)
        ymin, ymax = ax.get_ylim()
        ymaxlabel = int(np.ceil(ymax))
        #ax.set_yticks([ymax])
        ax.set_yticks([])
        #ax.set_yticklabels([str(ymaxlabel)], ha='left', va='top', fontsize=fontsize)
        ax.spines['right'].set_position(('axes', 1.01))
        ax.tick_params(
            axis='y',
            labelsize=fontsize,
            right=True,
            left=False,
            length=2,
            which='both',
            labelright=True,
            labelleft=False,
            direction='in',
            #size=fontsize,
        )
        ax.set_ylabel(gene, rotation=0, fontsize=fontsize, ha='right', va='bottom',
                      style=italicstyle)
        ax.yaxis.set_label_coords(-0.005, 0.1)
    ax.set_xlim(0, len(categories) * nbins)
    #ax.tick_params(axis='x', bottom=False, labelbottom=False)
    ax.set_xticks(nbins * (np.arange(len(categories))+0.5))
    ax.set_xticklabels(category_labels, ha='left', va='top', rotation=label_rotation,
                       fontsize=fontsize)
    # groupby_ax.tick_params(bottom=True)
    if type=="hotspot":
        ax.set_xlabel('Hotspot', fontsize=fontsize)
        fig.suptitle(' - '.join([v.capitalize() for v in interaction.split('_')]), fontsize=fontsize, y=0.95)

    # the ax to plot the groupby categories is split to add a small space between the rest of the plot and the categories
    #axs2 = gridspec.GridSpecFromSubplotSpec(
    #    2, 1, subplot_spec=axs[num_rows - 1, 0], height_ratios=[1, 1]
    #)

    #groupby_ax = fig.add_subplot(axs2[1])
    #groupby_ax.axis('off')

    #fig.tight_layout()
    return fig, ax


if __name__ == "__main__":
    import geospace
    import os
    available_datasets = ["V1_Mouse_Brain_Sagittal_Anterior", "V1_Mouse_Brain_Sagittal_Posterior",
                          "seqfish", "merfish",
                          "slideseq", "V1_Breast_Cancer_Block_A_Section_1"]
    dataset = "V1_Breast_Cancer_Block_A_Section_1"
    cache_dir = os.path.expanduser(f"~/Dropbox/data/ms/datasets/{dataset}")
    adata = geospace.read(os.path.join(cache_dir, 'adata'))
    markers = geospace.geometric_markers(adata, [0, 1, 2, 5])
    geospace.plot.tracks_plot(adata, markers)
    plt.show()


import os

import matplotlib.pyplot as plt
import scanpy as sc
import seaborn as sns
import numpy as np

# Global variables store dataset-specific information
from geospace.hotspot.geometry import similarity_map

library_id = None
img = None
scale = None
spot_size_preset = None
parameters_set = False
image_path = None


def set_dataset_plot_parameters(dataset):
    """
    Sets global variables that are used in spatial plots for non-Visium datasets
    :param dataset:
    :return:
    """
    global library_id, img, scale, spot_size_preset, parameters_set, image_path
    if dataset == "seqfish":
        img = None
        scale = None
        spot_size_preset = 0.045
    elif "merfish" in dataset:
        img = None
        scale = None
        spot_size_preset = 0.02
    elif "imc" in dataset:
        img = None
        scale = None
        spot_size_preset = 10
    elif "slideseq" in dataset:
        img = None
        scale = None
        spot_size_preset = 30
    else:
        pass

    # So that we can check that the globals are set correctly later
    parameters_set = True


def spatial(adata, spot_size=None, **kwargs):
    """
    Wraps scanpy's pl.spatial function, using parameters from set_dataset_plot_parameters,
    in order to make it a bit
    less annoying to manage parameters when analyzing non-Visium data
    :param adata: adata object to plot (spatial information in .obsm['spatial'])
    :param spot_size: size of spots used to show color parameter
    :param kwargs: Other keyword arguments to pass to sc.pl.spatial, such as color=, etc.
    :return:
    """

    if spot_size is None:
        spot_size = spot_size_preset

    sc.pl.spatial(adata, img=img, scale_factor=scale, spot_size=spot_size,
                  **kwargs)

    try:
        ax = kwargs['ax']
        ax.set_xlabel('')
        ax.set_ylabel('')
    except (KeyError, AttributeError):
        pass


def volcano(df, interaction=None, gene_list=None, ax=None, show=True, xlim=None, ylim=None,
            axis_labels=True, fc_cutoff=0, p_cutoff=2, fontsize=None, xoffset=0, yoffset=0.5,
            marker=None, title=None,
            **kwargs):
    if interaction is not None:
        # filter out the actual genes from the interaction
        ligand, receptor = interaction.split("_")
        ligand = ligand.capitalize()
        receptor = receptor.capitalize()
        df = df.drop(index=ligand).drop(index=receptor)
    x = df["log2(fc)"]
    y = df["-log10(p)"]

    # for coloring points based on significance
    hue = np.zeros(shape=x.shape)
    hue[np.logical_and(x > fc_cutoff, y > p_cutoff)] = 1
    hue[np.logical_and(x < -fc_cutoff, y > p_cutoff)] = 2
    palette = []
    if 0 in hue:
        palette.append(sns.color_palette("muted")[7])
        # palette.append("gray")
    if 1 in hue:
        palette.append(sns.color_palette("muted")[3])
        # palette.append("red")
    if 2 in hue:
        palette.append(sns.color_palette("muted")[0])
        # palette.append("blue")

    # For indicating a subset of genes using a different marker
    style = None
    if marker is not None:
        style = np.zeros(len(x), dtype=np.int32)
        for idx in range(len(x)):
            style[idx] = df.index[idx] in marker

    ax = sns.scatterplot(x=x, y=y, hue=hue, palette=palette, legend=None, ax=ax, style=style,
                         markers=['o', '^'],
                         **kwargs)

    for i, row in df.iterrows():
        x = row["log2(fc)"]
        y = row["-log10(p)"]
        if gene_list is not None and (gene_list is True or i in gene_list):
            # gene_list can be either a list or a dictionary containing offsets
            try:
                gene_xoffset, gene_yoffset = gene_list[i]
            except TypeError:
                gene_xoffset, gene_yoffset = xoffset, yoffset
            ax.text(x + gene_xoffset, y + gene_yoffset, i, fontsize=fontsize,
                    horizontalalignment='right')

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    if not axis_labels:
        ax.set_xlabel('')
        ax.set_ylabel('')
    else:
        ax.set_xlabel('log2(fc)', size=fontsize)
        ax.set_ylabel('-log10(p)', size=fontsize)
    plt.xticks(size=fontsize)
    plt.yticks(size=fontsize)

    # TODO: titles
    if title is not None:
        ax.set_title(title)
    elif interaction is not None:
        ax.set_title(" - ".join([v.capitalize() for v in interaction.split("_")]))
    if show:
        plt.show()

    return ax


def plot_similarity_map(adata, idx, ax=None, linewidth=0.5, linecolor="black", title="",
                        show=False, **kwargs):
    if ax is None:
        fig, ax = plt.subplots()
    try:
        scale_factor = list(adata.uns['spatial'].values())[0]['scalefactors'][
            'tissue_hires_scalef']
    except KeyError:
        if scale is not None:
            scale_factor = scale
        else:
            scale_factor = 1
    try:
        boundary = scale_factor * adata.uns["multi_boundaries"][str(idx)]
    except KeyError:
        raise ValueError("Need to compute boundaries first")
    #ax.fill(boundary[:, 0], boundary[:, 1], color_string)
    ax.plot(boundary[:, 0], boundary[:, 1], linewidth=linewidth, color=linecolor)

    arr = similarity_map(adata, idx=idx)

    adata.obs['tmp'] = arr
    spatial(adata, color="tmp", alpha_img=0.5, ax=ax, title=title, frameon=False, show=show,
            **kwargs)


from tqdm import tqdm
import numpy as np
import pandas as pd

from geospace.hotspot.hotspot import _compute_cutoff, compute_hotspots
from geospace.utility.utility import get_activity_dataframe
from geospace.cci.cci import compute_activity


def interaction_hotspots(adata, verbose=False, interactions=None, save_activity=False,
                         method='permutation', min_active_count=1, sig_threshold=0.95, **kwargs):
    # check if activity scores are available
    activity_matrix = get_activity_dataframe(adata)
    if len(list(activity_matrix)) == 0:
        # TODO: really should only compute activity for interactions in the filter if interactions
        # is not None
        try:
            um_scale = adata.uns['um_scale']
        except KeyError:
            # Try to get it automatically assuming Visium data with 55um spots
            try:
                dataset = list(adata.uns['spatial'].keys())[0]
                um_scale = adata.uns['spatial'][dataset]['scalefactors']['spot_diameter_fullres']/55
            except KeyError:
                raise ValueError("Please set adata.uns['um_scale'] to the length of one micrometer"
                                 " in the units of adata.obsm['spatial'].") from None

        secreted_std = 40*um_scale
        contact_threshold = 20*um_scale

        perform_permutations = method == 'permutation'
        activity_matrix = compute_activity(adata, secreted_std=secreted_std,
                                           contact_threshold=contact_threshold,
                                           sig_threshold=sig_threshold,
                                           perform_permutation=perform_permutations,
                                           save_activity=save_activity, verbose=verbose,
                                           min_active_count=min_active_count,
                                           interactions=interactions)

    cols = enumerate(list(activity_matrix))

    if verbose:
        print("Computing CCI hotspots")
        # everything precomputed so following step does not take significant time
        #cols = tqdm(cols, total=activity_matrix.shape[1])

    region_dict = {}

    for _, interaction in cols:
        if interactions is not None and interaction not in interactions:
            # interaction filter
            continue

        data = activity_matrix[interaction]

        if method == 'permutation':
            cutoff = adata.uns['activity_significance_cutoff'][interaction]
        else:
            cutoff = _compute_cutoff(data)

        inds = np.where(data > cutoff)[0]

        regions = compute_hotspots(adata=adata, input_data=inds, return_regions=True, **kwargs)

        if regions is not None:
            region_dict[f"hotspots_{interaction}"] = regions

    hotspots_df = pd.DataFrame(region_dict, index=adata.obs.index)
    adata.obs = pd.concat([adata.obs, hotspots_df], axis=1)



import warnings
import scipy.sparse

import numpy as np

from sklearn.neighbors import BallTree
from sklearn.preprocessing import normalize
from scipy.spatial import Delaunay


def get_neighbor_adjacency(coords, eps=None, k=None, weight=None, return_row_col=False):
    """
    Compute either k-nearest neighbor or epsilon-radius neighbor adjacency graph
    :param coords: Spatial coordinates of each cell
    :param eps: radius cutoff for epsilon-radius neighbors
    :param k: number of nearest neighbors for k-nearest neighbors
    :param weight: Function mapping distance to value in weighted adjacency (default: f(x) = 1)
    :param return_row_col: If true, return the row and column variables from sparse coo format
    :return:
    """
    tree = BallTree(coords)
    n_cells = len(coords)
    if eps is not None and k is not None:
        warnings.warn('k is ignored if eps is provided')

    if eps is not None:
        nearby_col, dist = tree.query_radius(coords, eps, return_distance=True)
        coords = []
        nearby_row = []
        for i, col in enumerate(nearby_col):
            nearby_col[i] = col
            if weight is not None:
                coords.append(weight(dist[i]))
            nearby_row.append(i * np.ones(shape=col.shape, dtype=col.dtype))
    elif k is not None:
        nearby_col = tree.query(coords, k,
                                return_distance=False)  # [:, 1:]  # remove self connections
        nearby_row = []
        for i, col in enumerate(nearby_col):
            nearby_row.append(i * np.ones(shape=col.shape, dtype=col.dtype))
    else:
        raise ValueError("Provide either distance eps or number of neighbors k")

    nearby_row = np.concatenate(nearby_row)
    nearby_col = np.concatenate(nearby_col)
    if weight is None or k is not None:
        coords = np.ones(shape=nearby_row.shape)
    else:
        coords = np.concatenate(coords)

    nearby = scipy.sparse.csr_matrix((coords, (nearby_row, nearby_col)),
                                     shape=(n_cells, n_cells))

    if return_row_col:
        return nearby, nearby_row, nearby_col

    return nearby


def compute_spatial_transport_matrices(adata,
                                       secreted_std,
                                       secreted_threshold_cutoff,
                                       contact_threshold=None):
    """
    Return the spatial prior belief of possibility of interaction between every possible pair of
    cells for both secreted signaling and cell-cell contact type interactions
    """

    ncells = adata.shape[0]

    # gaussian diffusion kernel - threshold is two standard deviations
    sigma = secreted_std

    def weight_fn(x):
        return np.exp(-0.5 * ((x / (2 * sigma)) ** 2))

    coords = adata.obsm['spatial']
    transport_secreted = get_neighbor_adjacency(coords, eps=sigma * secreted_threshold_cutoff,
                                                weight=weight_fn)

    # delauney for direct contact
    tri = Delaunay(adata.obsm['spatial'])
    indptr, indices = tri.vertex_neighbor_vertices
    data = np.ones(shape=indices.shape)
    transport_contact = scipy.sparse.csr_matrix((data, indices, indptr), shape=[ncells, ncells])

    # filter out edges in delauney triangulation that are too long
    if contact_threshold is not None:
        transport_contact = transport_contact.multiply(
            get_neighbor_adjacency(coords, eps=contact_threshold))

    # normalize across number of neighbors
    transport_contact = normalize(transport_contact, norm='l1', axis=1)
    transport_secreted = normalize(transport_secreted, norm='l1', axis=1)

    adata.obsp['transport_contact'] = transport_contact
    adata.obsp['transport_secreted'] = transport_secreted
